"""Provides utilities for scanning the music library filesystem for files.

This module provides a LibraryScanner class that can be used to walk through a directory and
pick up any files relevant to musicbird, plus changes made to it.
"""

import logging
from pathlib import Path

from .db import LibraryDB
from .file import File


logger = logging.getLogger(__name__)

# Always ignore these files as system metadata
IGNORE_FILES = [
    "Thumbs.db"
]


class LibraryScanner:
    """Index a directory and add its contents to the DB.

    Use this class to scan a directory for files and add them to the provided Database.
    """

    def __init__(self, path: Path, db: LibraryDB) -> None:
        """Generate a new scanner for the given path and database.

        Args:
            path (Path): The path to scan.
            db (LibraryDB): The library to save the result to.
        """
        self.path = path.resolve()
        self.db = db

    def scan(self) -> bool:
        """ Scan the filesystem for changes.

        Registers new, modified and deleted files and updates the Library.

        Returns:
            bool: True if all files were scanned successfully, False if not.
        """
        failed = False

        # Mark all files as deleted in-memory at the beginning of our scan
        # As files are rediscovered, the deleted flag will be removed
        logger.info(f"Scanning directory {self.path} into library")

        for file in self.db.get_all_files():
            file.was_deleted = True
            self.db.add_or_update_file(file)

        for path in self.path.glob("./**/*"):
            if path.is_file() and path.name not in IGNORE_FILES:
                file = File(path)
                if not self.add_or_update_file(file):
                    failed = True

        return not failed

    def add_or_update_file(self, file: File) -> bool:
        """Adds a single new file to the library or update an existing one.

        Args:
            file (File): The file object to add

        Returns:
            bool: True if the file was processed successfully, False if not
        """
        # If we're adding/updating a file, that means it must also exist physically
        file.was_deleted = False

        current_entry = self.db.get_file_by_path(file.path)

        # By only scanning for the filetype after we detected a new/changed file, we can save a lot of calls
        # to ffmpeg, thus considerably speeding up scanning. If the file is unchanged, we just reuse the old type.
        if not current_entry:
            logger.info(f"Adding new file to library: {file.path}")
            file.determine_type()
            file.needs_processing = True
        elif current_entry.mtime != file.mtime:
            logger.info(f"Existing file has been modified and will be reprocessed: {file.path}")
            file.determine_type()
            file.needs_processing = True
        else:
            logger.debug(f"File unchanged since last scan: {file.path}")
            file.type = current_entry.type
        self.db.add_or_update_file(file)
        return True



"""Provides the scan processing step and related functions.
"""

import argparse
import logging
from typing import Dict, List

from .db import LibraryDB, init as init_db
from .scanner import LibraryScanner

logger = logging.getLogger(__name__)


def scan_command(parent_parser: argparse.ArgumentParser, args: List[str], config: Dict) -> bool:
    """Entrypoint for the CLI `scan` command.

    Args:
        parent_parser (argparse.ArgumentParser): The parser from the main entrypoint.
            Used to display a full --help output by inheriting its arguments.
        args (List[str]): List of arguments not parsed by the main parser.
        config (Dict): Dictionary containing the MusicBird configuration
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=__doc__, prog="musicbird", parents=[parent_parser])
    parser.add_argument("--pretend", action="store_true", help="Show scan results but don't store them")
    args = parser.parse_args(args)
    return scan(config, init_db(config, pretend=args.pretend))


def scan(config: Dict, db: LibraryDB) -> bool:
    """Scan the source library for changed files and write them to the DB.

    Performs a filesystem scan on the source music library, registering any new, changed or deleted files along the way.
    Once files are registered, they are then stored in the Database for usage by other commands.

    Args:
        config (Dict): Dictionary containing the musicbird configuration.
        db (LibraryDB): Database object to read/write the library status from/to.
        pretend (bool, optional): Pretend to scan, but don't modify the DB. Defaults to False.

    Returns:
        bool: True if the scan was successful, False if not.
    """
    scanner = LibraryScanner(config["source"], db)
    logger.info("Scanning library...")
    result = scanner.scan()
    logger.info((
        f"Scanned library. Summary: {len(db.get_all_files())} total files, {len(db.get_files_needing_processing())} "
        f"files to process, {len(db.get_deleted_files())} files deleted."
    ))
    return result



#!/usr/bin/env python3

import argparse
import logging
from pathlib import Path
import sys
from typing import List

from schema import SchemaError

from . import config, run, scan, prune, copy, encode, __version__

logger = logging.getLogger("musicbird")


def main(args: List[str]) -> bool:
    """Main command-line entrypoint.

    Called from any of the various ways to start MusicBird. Parses general arguments
    and initializes some basics (configuration), before handing off control to the
    called subcommands `$commandname_command` method

    Args:
        args (list): A list of command-line arguments, excluding argv[0] (the call to "musicbird" itself).

    Returns:
        bool: True if all commands completed successfully, False if not.
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False, prog="musicbird")
    parser.add_argument("-c", "--config", help="Path to the configuration file to use",
                        default=config.Config.DEFAULT_PATH)
    parser.add_argument("--loglevel", "-v", "--verbosity", help="Set the output verbosity",
                        choices=["DEBUG", "INFO", "WARNING", "ERROR", "FATAL"], default="INFO")
    parser.add_argument("--version", help="Print the program version and exit", action="store_true")
    parser.add_argument("command", nargs="?", help="The command you want to run", choices=[
                        "config", "run", "scan", "copy", "encode", "prune"])
    args, command_args = parser.parse_known_args(args)

    logging.basicConfig(level=getattr(logging, args.loglevel))

    if args.version:
        print(__version__)
        return True

    config_path = Path(args.config)
    # The "config" command needs to be runnable without having loaded an existing config.
    if args.command not in ["config"]:
        try:
            _config = config.Config(config_path)
        except OSError:
            sys.exit("Could not read config file. You can generate a new config file with 'musicbird config generate'")
        except SchemaError:
            sys.exit("Your configuration file is invalid. Please check the error message above for details")

    if args.command == "config":
        successful = config.config_command(parser, command_args, config_path)
    elif args.command == "run":
        successful = run.run_command(parser, command_args, _config.config)
    elif args.command == "scan":
        successful = scan.scan_command(parser, command_args, _config.config)
    elif args.command == "copy":
        successful = copy.copy_command(parser, command_args, _config.config)
    elif args.command == "encode":
        successful = encode.encode_command(parser, command_args, _config.config)
    elif args.command == "prune":
        successful = prune.prune_command(parser, command_args, _config.config)
    else:
        parser.parse_args()
        successful = False

    if not successful:
        return False
    return True


def entrypoint():
    """Pre-Entrypoint for the program.

    Used to pass arguments to the main function explicitly as this makes testing
    command line parameters a lot easier.
    """
    if not main(sys.argv[1:]):
        sys.exit("Some commands did not complete successfully")


if __name__ == "__main__":
    entrypoint()



"""Provides the encode processing step and related functions.
"""

import argparse
import concurrent.futures
import logging
from queue import Queue
import threading
import time
from typing import Dict, List

from .db import LibraryDB, init as init_db
from .file import File, FileType

logger = logging.getLogger(__name__)


def encode_command(parent_parser: argparse.ArgumentParser, args, config: Dict) -> bool:
    """Entrypoint for the CLI `encode` command.

    Args:
        parent_parser (argparse.ArgumentParser): The parser from the main entrypoint.
            Used to display a full --help output by inheriting its arguments.
        args (List[str]): List of arguments not parsed by the main parser.
        config (Dict): Dictionary containing the MusicBird configuration
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=__doc__, prog="musicbird", parents=[parent_parser])
    parser.add_argument("--pretend", action="store_true",
                        help="Show what files would be converted, but don't perform the actual encoding")
    args = parser.parse_args(args)
    db = init_db(config, pretend=args.pretend)
    return encode(config, db, args.pretend)


def _encode_worker(file: File, config: Dict, successes: Queue, failures: Queue) -> bool:
    """Thread function that processes a single file, then returns.

    Called by encode(), this worker first encodes its file,
    then puts the file into either the successes or failures queue, depending on whether everything went well.

    Args:
        file (File): The file to encode.
        config (Dict): MusicBird config dict.
        successes (Queue): Queue for successfully encoded files.
        failures (Queue): Queue for failed encodes.

    Returns:
        bool: True if the encode was successful, False if not
    """
    if file.encode_to_dest(config):
        file.needs_processing = False
        successes.put(file)
        return True
    else:
        failures.put(file)
        return False


def _db_update_worker(db: LibraryDB, input_q: Queue, output: Queue) -> None:
    """Processes encoded files and saves them to the DB.

    Processes files will also be put into the output queue.
    Terminates upon receiving a None object.

    args:
        db (LibraryLB): Database to write to.
        input_q (Queue): Queue to read files from.
        output (Queue): Queue to send processed files to.
    """
    while True:
        if not input_q.empty():
            file = input_q.get()
            if not file:
                # None object was sent by controll thread -> exit
                return
            db.add_or_update_file(file)
            output.put(file)
            logger.info(f"Processed file: {file.path}")
        else:
            time.sleep(1)


def encode(config: Dict, db: LibraryDB, pretend=False) -> bool:
    """Process and copy all files marked for copying to the mirror library.

    Encodes all files that are due for encoding, according to the data in the library DB.
    Will encode lossles files and also lossy files, if the configuration option is set accordingly.

    Args:
        config (Dict): Dictionary containing the musicbird configuration
        db (LibraryDB): Database object to read/write the library status from/to.
        pretend (bool, optional): Pretend to encode, but don't perform any actual operations. Defaults to False.

    Returns:
        bool: True if all files were processed successfully, false if not.
    """
    to_encode: List[File] = []
    for file in db.get_files_needing_processing():
        if file.type == FileType.LOSSLESS or (file.type == FileType.LOSSY and config["lossy_files"] == "convert"):
            to_encode.append(file)
        # Remove the processing flag from lossy files if they're to be ignored
        if file.type == FileType.LOSSY and config["lossy_files"] == "ignore":
            file.needs_processing = False
    logger.info(f"Need to encode {len(to_encode)} files")

    if not pretend:
        # Queue structure:
        # encode_worker --- success ---> successful_encodes ---> db_update_worker ---> processed_files
        #      |
        #      +----------- failure ---> failed_encodes
        successful_encodes = Queue()
        failed_encodes = Queue()
        processed_files = Queue()

        db_thread = threading.Thread(target=_db_update_worker, args=(db, successful_encodes, processed_files))
        db_thread.start()

        with concurrent.futures.ThreadPoolExecutor(config["threads"]) as executor:
            for file in to_encode:
                executor.submit(_encode_worker, file, config, successful_encodes, failed_encodes)

        # Encode jobs have finished, add termination object to DB queue
        successful_encodes.put(None)
        db_thread.join()

        # Process failures
        failures: List[File] = []
        while not failed_encodes.empty():
            failures.append(failed_encodes.get_nowait())

    else:
        processed_files = []
        for file in to_encode:
            file.needs_processing = False
            db.add_or_update_file(file)
            processed_files.append(file)
            logger.info(f"Encoded file: {file.path}")
        failures = []

    logger.info(f"Successfully encoded {len(to_encode) - len(failures)} files")
    if failures:
        logger.error(f"Failed to delete {len(failures)} files. See above for errors")
        failures_str = "\n".join([str(f.path) for f in failures])
        logger.debug(f"Failed files: {failures_str}")
        return False
    else:
        return True



"""Provides a representation/abstraction layer for files in the underlying Filesystem.

This modules main purpose is to provide the File class, which represents a file in the source music library.
All file-based operations should be performed via File's methods.
"""

from enum import Enum
import logging
import os
from pathlib import Path
from shutil import copy
from typing import Dict

import ffmpeg

from .encoder import init as init_encoder

logger = logging.getLogger(__name__)

LOSSY_CODECS = [
    "mp3",
    "opus",
    "ogg",
    "wmav1",
    "wmav2",
    "aac"
]
LOSSLESS_CODECS = [
    "flac",
    "wav",
    "aiff",
]

ALBUMART_FILENAMES = [
    "cover",
    "album",
    "albumart",
]
ALBUMART_EXTENSIONS = [
    ".jpg",
    ".jpeg",
    ".png",
]


class FileType(Enum):
    """Enum used to set the type of file, as far as MusicBird is concerned.

    In order to process various kinds files, each File has a specific type
    attached to it. This enum provides all possible values for that attribute.
    """

    LOSSY = 1
    """A lossy audio file(MP3, Opus, WAV, ...)"""

    LOSSLESS = 2
    """A lossless audio file(FLAC, WAV, AIFF)"""

    ALBUMART = 3
    """An image file that's most likely the album art cover"""

    OTHER = 4
    """Any other kind of file"""


class File:
    """Provides an abstraction layer for all operations on the music library source files.

    A File object represents one file in the original music library. File is used used
    to store information about the files relevant attributes as well as to provide
    methods for processing each file. File objects are usually generated during scan
    operations and are then stored in the Database, from where other actions can retrieve and process them.

    File objects can be compared with each other, and are sortable(using the path as a sorting base).
    Since they are mutable, they are not hashable and cannot be used as keys in dictionaries, among other things.

    Attributes:
        path: Path object containing the full filesystem Path to the underlying file.
        filetype: Kind of file, represented by a FileType enum.
        mtime: Last time the file was modified. Stored as a UNIX timestamp(Integer).
        needs_processing: Bool indicating whether the file requires processing.
            This usually means that file was modified/added recently.
        was_deleted: Bool indicating whether the file was deleted and no longer exists on the fs.
            Used to track deletions.
    """

    def __init__(self, path: Path, filetype: FileType = None, mtime: int = None,
                 needs_processing: bool = False, was_deleted: bool = False) -> None:
        """Creates a new File object, representing a physical file in the source libary.

        Args:
            path(Path): Path object containing the full filesystem Path to the underlying file.
            filetype(FileType, optional): The kind of the file(lossless/lossy/...). Use the FileType enum for this.
            mtime(int, optional): Last time the file was modified as a UNIX timestamp(int).
                Will read the current mtime if not specified.
            needs_processing(bool, optional): Bool indicating whether the file requires processing.
                This usually means that file was modified/added recently.
            was_deleted(bool, optional): Bool indicating whether the file was deleted and no longer exists on the fs.
                Used to track deletions.
        """
        self.path = path
        self.needs_processing = needs_processing
        self.was_deleted = was_deleted
        self.type = filetype

        if not mtime:
            self.mtime = round(os.path.getmtime(path))
        else:
            self.mtime = mtime

    def copy_to_dest(self, config: Dict) -> bool:
        """Copy the file to the destination library.

        Copies the file to its relative path in the destination library specified in config.
        For example, if the original file is in ~/music/Artist1/Album1/Track1.mp3 and the
        destination directory is ~/music_converted, then the file will be copied to
        ~/music_converted/Artist1/Album1/Track1.mp3. Any missing directories will be created.

        Args:
            config(dict): Musicbird config as a dict.

        Returns:
            bool: True if the copy operation was successful, False if not.
        """
        if self.was_deleted:
            logger.error(f"File {self.path} was deleted and cannot be copied")
            self.was_deleted = True
            return False

        dest = self.get_dest_path(config)
        logger.info(f"Copying file: {self.path}")
        try:
            dest.parent.mkdir(parents=True, exist_ok=True)
            logger.debug(f"Copying to: {dest}")
            copy(self.path, dest)
        except OSError as e:
            logger.error(f"Could not copy file {self.path} to {dest}: {repr(e)}")
            return False
        return True

    def encode_to_dest(self, config: Dict) -> bool:
        """Encode the file and save it in the destination library.

        Encodes the file to its relative path in the destination library specified in config.
        For example, if the original file is in ~/music/Artist1/Album1/Track1.flac and the
        destination directory is ~/music_converted, then the file will be encoded to
        ~/music_converted/Artist1/Album1/Track1.mp3/opus/...

        Any missing directories will be created.

        Args:
            config(dict): Musicbird config as a dict.

        Returns:
            bool: True if the encode operation was successful, False if not.
        """
        dest = self.get_dest_path(config)
        encoder = init_encoder(config)
        return encoder.encode(self.path, dest)

    def get_dest_path(self, config: Dict) -> Path:
        """Get the files path in the destination library.

        Generates the path for the mirror copy of this file in the destination library, as specified by the config.
        For example, if the original file is in ~/music/Artist1/Album1/Track1.mp3 and the
        destination directory is ~/music_converted, then the destination path would be
        ~/music_converted/Artist1/Album1/Track1.mp3.

        If the file due to be encoded, the file extension is also changed to reflect this(e.g. .flac -> .mp3)

        Args:
            config(dict): The musicbird configuration in dict form.

        Returns:
            bool: A Path object pointing to the path of the destination file.
        """
        src = config["source"]
        dest = config["destination"]

        if not self.type:
            self.determine_type()

        path = Path(str(self.path).replace(str(src), str(dest), 1))
        if self.type == FileType.LOSSLESS or (self.type == FileType.LOSSY and config["lossy_files"] == "convert"):
            path = path.with_suffix(init_encoder(config).extension)
        return path

    def determine_type(self) -> None:
        """Detects the type of file and sets the filetype attribute accordingly.

        Tries to figure out the kind of file present at self.path based on a few tests,
        then sets the filetype attribute to the best match. Among other things, it uses
        ffmpeg to parse media files and checks the filename for patterns(such as cover.jpg) files.

        Since this does actually need to physically access the file, it is not called during File
        initialization. This method is especially useful if you are adding a new file and don't know its exact type yet.
        """
        if self.path.name.lower() in [name + ext for name in ALBUMART_FILENAMES for ext in ALBUMART_EXTENSIONS]:
            self.type = FileType.ALBUMART
            return

        try:
            probe = ffmpeg.probe(self.path)
        except (OSError, ffmpeg.Error) as e:
            if "Invalid data found when processing input" in str(e.stderr):
                # ffmpeg can't handle the file, so its safe to assume that it's something else. Binary, text, whatever
                self.type = FileType.OTHER
                return
            else:
                logger.warning(
                    f"Failure trying to determine type for file {self.path}, falling back to type 'OTHER': {repr(e)}.")
                self.type = FileType.OTHER
                return

        # Generate a set of audio codecs used in the file. Regular audio files usually only have a single stream.
        audio_codecs = list({probe["streams"][i]["codec_name"] for i in range(len(probe["streams"]))
                            if probe["streams"][i]["codec_type"] == "audio"})

        if not audio_codecs:
            self.type = FileType.OTHER  # Some other non-audio file recognized by FFMPEG
        elif len(audio_codecs) > 1:
            logger.warning(f"File has multiple audio streams: {self.path}. Cannot encode safely, Will copy instead.")
            self.type = FileType.OTHER
        elif audio_codecs[0] in LOSSLESS_CODECS:
            self.type = FileType.LOSSLESS
        elif audio_codecs[0] in LOSSY_CODECS:
            self.type = FileType.LOSSY
        else:
            self.type = FileType.OTHER  # Audio file that we don't recognize. Copy instead

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, File):
            return False
        return (self.path == o.path
                and self.type == o.type
                and self.mtime == o.mtime
                and self.was_deleted == o.was_deleted
                and self.needs_processing == o.needs_processing
                )

    def __lt__(self, o: object) -> bool:
        return self.path < o.path

    def __gt__(self, o: object) -> bool:
        return self.path > o.path



"""Provides the copying processing step and related functions.
"""

import argparse
import logging
from typing import Dict, List

from .db import LibraryDB, init as init_db
from .file import File, FileType

logger = logging.getLogger(__name__)


def copy_command(parent_parser: argparse.ArgumentParser, args: List[str], config: Dict) -> bool:
    """Entrypoint for the CLI `copy` command.

    Args:
        parent_parser (argparse.ArgumentParser): The parser from the main entrypoint.
            Used to display a full --help output by inheriting its arguments.
        args (List[str]): List of arguments not parsed by the main parser.
        config (Dict): Dictionary containing the MusicBird configuration
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=__doc__, prog="musicbird", parents=[parent_parser])
    parser.add_argument("--pretend", action="store_true",
                        help="Show what files would be copied, but don't perform the actual copy")
    args = parser.parse_args(args)
    db = init_db(config, pretend=args.pretend)
    return copy(config, db, args.pretend)


def copy(config: Dict, db: LibraryDB, pretend=False) -> bool:
    """Process and copy all files marked for copying to the mirror library.

    Processes all regular/copyable files according to their state in the library DB.
    Will copy regular files, album art and lossy files depending on the values set in config,
    before marking them as processed in the Databse.

    Args:
        config (Dict): Dictionary containing the musicbird configuration.
        db (LibraryDB): Database object to read/write the library status from/to.
        pretend (bool, optional): Pretend to copy, but don't perform any filesystem operations. Defaults to False.

    Returns:
        bool: True if all files were processed successfully, false if not.
    """
    to_copy: List[File] = []
    for file in db.get_files_needing_processing():
        if ((file.type == FileType.OTHER and config["copy"]["files"])
            or (file.type == FileType.ALBUMART and config["copy"]["album_art"])
            or (file.type == FileType.LOSSY and config["lossy_files"] == "copy")
            ):
            to_copy.append(file)
        # Remove the processing flag from files concerning us if the required option is not set
        if ((file.type == FileType.OTHER and not config["copy"]["files"])
            or (file.type == FileType.ALBUMART and not config["copy"]["album_art"])
            or (file.type == FileType.LOSSY and config["lossy_files"] == "ignore")
            ):
            file.needs_processing = False
    logger.info(f"Need to copy {len(to_copy)} files")

    successes = []
    failures = []
    for file in to_copy:
        if not pretend:
            if file.copy_to_dest(config):
                file.needs_processing = False
                db.add_or_update_file(file)
                successes.append(file)
            else:
                failures.append(file)
        else:
            file.needs_processing = False
            db.add_or_update_file(file)
            successes.append(file)

    logger.info(f"Successfully copied {len(successes)} files")
    if failures:
        logger.error(f"Failed to copy {len(failures)} files. See above for errors")
        failures_str = "\n".join([str(f.path) for f in failures])
        logger.debug(f"Failed files: {failures_str}")
        return False
    else:
        return True



"""Provides the prune processing step and related functions.
"""

import argparse
import os
import logging
from typing import Dict, List

from .db import LibraryDB, init as init_db
from .file import File

logger = logging.getLogger(__name__)


def prune_command(parent_parser: argparse.ArgumentParser, args: List[str], config: Dict) -> bool:
    """Entrypoint for the CLI `prune` command.

    Args:
        parent_parser (argparse.ArgumentParser): The parser from the main entrypoint.
            Used to display a full --help output by inheriting its arguments.
        args (List[str]): List of arguments not parsed by the main parser.
        config (Dict): Dictionary containing the MusicBird configuration
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=__doc__, prog="musicbird", parents=[parent_parser])
    parser.add_argument("--pretend", action="store_true",
                        help="Show what files would be pruned, but don't perform the actual deletion")
    args = parser.parse_args(args)
    db = init_db(config, pretend=args.pretend)
    return prune(config, db, args.pretend)


def prune(config: Dict, db: LibraryDB, pretend=False) -> bool:
    """Process and delete all files marked for deletion in the mirror library.

    Attempts to delete all files that are marked as deleted in the mirror library,
    then marks them as processed.

    Args:
        config (Dict): Dictionary containing the musicbird configuration.
        db (LibraryDB): Database object to read/write the library status from/to.
        pretend (bool, optional): Pretend to delete, but don't perform any filesystem operations. Defaults to False.

    Returns:
        bool: True if all files were processed successfully, false if not.
    """
    to_prune = db.get_deleted_files()
    logger.info(f"Need to delete {len(to_prune)} files")

    successes: List[File] = []
    failures: List[File] = []
    for file in to_prune:
        dest = file.get_dest_path(config)
        if not pretend:
            try:
                dest.unlink()
            except FileNotFoundError:
                db.remove_file(file)
                successes.append(file)
                logger.info(f"Removed file: {dest}")
            except OSError as e:
                logger.error(f"Could not remove file {dest}: {repr(e)}")
                failures.append(file)
            else:
                db.remove_file(file)
                successes.append(file)
                logger.info(f"Removed file: {dest}")
        else:
            successes.append(file)
            db.remove_file(file)

        if not os.listdir(dest.parent):
            # Remove empty leftover directories
            try:
                os.removedirs(dest.parent)
            except FileNotFoundError:
                pass
            except OSError as e:
                logger.warning(f"Could not remove empty directory {dest.parent}: {repr(e)}")

    logger.info(f"Successfully pruned {len(successes)} files")
    if failures:
        logger.error(f"Failed to delete {len(failures)} files. See above for errors")
        failures_str = "\n".join([str(f.path) for f in failures])
        logger.debug(f"Failed files: {failures_str}")
        return False
    else:
        return True



"""Tools related to reading and parsing the MusicBird configuration file.

This module provides utilities to interface with MusicBirds configuration file,
as well as the entrypoint for the config CLI command.
"""

import argparse
import logging
import json
import re
import os
from pathlib import Path
import sys
from typing import Dict, List, Union

from pkg_resources import resource_stream
from schema import Schema, And, SchemaError, Use, Optional
import yaml

logger = logging.getLogger(__name__)


class Config:
    """Class representing the MusicBird configuration file.

    When created, this class loads a configuration file from disk, verifies that it is valid
    and does some basic processing, such as fully expanding all file paths. You can then access
    the configuration using the .config attribute.

    Attributes:
        config: Configuration, as a dict
    """

    _SCHEMA = Schema({
        "source": And(Use(Path)),
        "database": And(Use(str), len, lambda d: d in ("sqlite3")),
        Optional("sqlite3"): {
            "path": And(Use(Path))
        },
        "destination": And(Use(Path)),
        "copy": {
            "files": And(Use(bool)),
            "album_art": And(Use(bool))
        },
        "prune": And(Use(bool)),
        "lossy_files": And(Use(str), len, lambda l: l in ("copy", "convert", "ignore")),
        "encoder": And(Use(str), len, lambda f: f in ("mp3", "opus")),
        "mp3": {
            "vbr": And(Use(bool)),
            Optional("quality"): And(Use(int), lambda q: 0 <= q <= 9),
            Optional("bitrate"): And(Use(str), lambda b: re.match(r'\d{1,4}k', b))
        },
        "opus": {
            "bitrate": And(Use(str), lambda b: re.match(r'\d{1,4}k', b))
        },
        "threads": And(Use(int), lambda t: t > 0)
    })
    # Name of the directory used to storing files related to this app
    _DIRNAME = "musicbird"
    _DEFAULT_CONFIG = {
        "database": "sqlite3",
        "sqlite3": {
            "path": f"{os.environ.get('XDG_DATA_HOME', os.environ['HOME'] + '/.local/share')}/{_DIRNAME}/db.sqlite3"
        },
        "copy": {
            "files": True,
            "album_art": False,
        },
        "prune": True,
        "lossy_files": "copy",
        "encoder": "mp3",
        "mp3": {
            "vbr": True,
            "quality": 0,
        },
        "opus": {
            "bitrate": "128k",
        },
        "threads": os.cpu_count()
    }

    DEFAULT_PATH = Path(
        f"{os.environ.get('XDG_CONFIG_HOME', os.environ['HOME'] + '/.config')}/{_DIRNAME}/config.yml")

    def __init__(self, path: Path) -> None:
        """Read a configuration file and parse/validate it.

        This will read a configuration file from disk, and perform some validation on it,
        before representing the result as part of the returned Object.

        Args:
            path (Path): The path to the config file on disk

        Raises:
            OSError: If the file could not be read
            schema.Error: If the configuration file is invalid
        """
        try:
            with path.open() as f:
                configfile = yaml.safe_load(f)
        except OSError as e:
            logger.fatal(f"Error while attempting to read config at {repr(e)}")
            raise
        if not configfile:
            logger.fatal(f"Empty config file at {path}")
            raise EOFError(f"Empty config file at {path}")

        self.config = Config._add_default_config(configfile)
        self._validate()
        self._resolve_paths(self.config)

    def print(self, output_format: str = "yml"):
        """Print the current configuration to console.

        Args:
            output_format (str, optional): Format to print the config in. Can be "yml" or "json". Defaults to "yml".
        """
        printable_config = Config._printify_paths(self.config)

        if output_format == "yml":
            print(yaml.safe_dump(printable_config))
        elif output_format == "json":
            print(json.dumps(printable_config))

    def _validate(self):
        """Validate the configuration file.

        Performs validation of the config file against the builtin schema.
        This might not catch every possible error, but it's a good start.

        Raises:
            SchemaError: Upon failed validation
        """
        try:
            self.config = Config._SCHEMA.validate(self.config)
        except SchemaError as e:
            logger.fatal(f"Invalid configuration file: {repr(e)}")
            raise e

    @staticmethod
    def _printify_paths(item: Union[Dict, List]) -> Union[Dict, List]:
        """Convert all path objects to strings.

        Converts all Path objects in item to strings, for usage with yaml/json dump.
        Calls itself recursively

        Args:
            item (Union[Dict, List]): The collection containing Path objects

        Returns:
            Union[Dict, List]: The collection with all paths converted
        """
        if isinstance(item, dict):
            for key in item:
                if isinstance(item[key], (dict, list)):
                    item[key] = Config._printify_paths(item[key])
                elif isinstance(item[key], Path):
                    item[key] = str(item[key])
        if isinstance(item, list):
            for i in range(len(item)):
                if isinstance(item[i], (dict, list)):
                    item[i] = Config._printify_paths(item[i])
                elif isinstance(item[i], Path):
                    item[i] = str(item[i])
        return item

    @staticmethod
    def _resolve_paths(item: Union[List, Dict]) -> Union[List, Dict]:
        """Make all paths in a collection absolute.

        This calls resolve() for all paths in a given collection, converting them to their absolute versions.
        Calls itself recursively.

        Args:
            item (Union[List, Dict]): The collection containing the non-absolute paths.

        Returns:
            Union[List, Dict]: The collection with all Paths resolved.
        """
        if isinstance(item, dict):
            for key in item:
                if isinstance(item[key], (dict, list)):
                    item[key] = Config._resolve_paths(item[key])
                elif isinstance(item[key], Path):
                    item[key] = item[key].expanduser().resolve()
        if isinstance(item, list):
            for i in range(len(item)):
                if isinstance(item[i], (dict, list)):
                    item[i] = Config._resolve_paths(item[i])
                elif isinstance(item[i], Path):
                    item[i] = item[i].expanduser().resolve()
        return item

    @staticmethod
    def _add_default_config(config: Dict) -> Dict:
        """Combine the default and concrete documentation into a single dict.

        Args:
            config (Dict): The concrete configuration on which to act.

        Returns:
            Dict: A dict containing the merged config.
        """
        merged_config = Config._recursive_merge(dict(Config._DEFAULT_CONFIG), config)
        return merged_config

    @staticmethod
    def _recursive_merge(a: Dict, b: Dict) -> Dict:
        """Merge two dictionaries

        Helper method that merges a dictionary (b) into another one (a).
        Modifies A.

        Args:
            a (Dict): The dictionary to merge into.
            b (Dict): The dictionary to add to a.

        Returns:
            Dict: The fully merged dictionary (a).
        """
        for key in b:
            if key in a:
                if isinstance(a[key], dict) and isinstance(b[key], dict):
                    Config._recursive_merge(a[key], b[key])
                else:
                    a[key] = b[key]
            else:
                a[key] = b[key]
        return a


def config_command(parent_parser: argparse.ArgumentParser, args: List[str], config_path: Path) -> bool:
    """Entrypoint for the CLI `config` command.

    This command is not part of the usual run lifecycle,
    but is instead a convenience command to create or view a configuration file.

    Current sub-actions are:
    * generate - To create a new config file
    * print - To print an existing config file

    Args:
        parent_parser (argparse.ArgumentParser): The parser from the main entrypoint.
            Used to display a full --help output by inheriting its arguments.
        args (List[str]): List of arguments not parsed by the main parser.
        config_path (Path): Path to the configuration file to use. In other CLI entrypoints, this would be the
            parsed config Object, but we can't do that here as we might be generating one for the first time.

    Returns:
        bool: True if the command was successful, False if not
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=__doc__, prog="musicbird", parents=[parent_parser])
    subparsers = parser.add_subparsers(dest="action")

    parser_generate = subparsers.add_parser("generate", help="Generate a new configuration file")
    parser_generate.add_argument("--output", "-o", default=Config.DEFAULT_PATH,
                                 help="Select the path at which to create the configfile")
    parser_generate.add_argument(
        "--force", "-f", help="Overwrite an already existing configuration file", action="store_true")

    parser_print = subparsers.add_parser("print", help="Print the currently loaded configuration")
    parser_print.add_argument("--format", "-f", default="yml", choices=["json", "yml"],
                              help="Print the configuration in this format")

    args = parser.parse_args(args)

    if args.action == "generate":
        output = Path(args.output)
        if output.exists() and not args.force:
            sys.exit(f"Configuration file at {output} already exists")

        config = resource_stream(__name__, "data/default_config.yml").readlines()
        try:
            output.parent.mkdir(exist_ok=True, parents=True)
            with output.open(mode='wb') as f:
                f.writelines(config)
        except OSError as e:
            sys.exit(f"Error while writing configuration: {repr(e)}")
        logger.info(f"Initialized configuration file at: {output}")
        return True

    elif args.action == "print":
        try:
            Config(config_path).print(args.format)
            return True
        except OSError:
            sys.exit("Could not read config file")
        except SchemaError:
            sys.exit("Your configuration file is invalid. Please check the error message above for details")



"""Database Abstractions and utilities.

This module provides the interface used by the other MusicBird components to read/write from a persistent Database.
The Abstract class LibraryLB serves as the interface used by other components, while individual DB implementations
(such as SQLiteLibrary) are derived from it.
"""

from abc import ABC, abstractmethod
import logging
from pathlib import Path
import sqlite3
from sqlite3.dbapi2 import Row
import sys
from typing import Dict, List, Tuple, Union

from .file import File, FileType

logger = logging.getLogger(__name__)


class LibraryDB(ABC):
    """Interface for communicating with a Database backend.

    This class defines all methods that clients can use to access the underlying Database.

    Note that the File path is used as a unique key, so there may only be one file for each path.
    """

    @abstractmethod
    def get_all_files(self) -> List[File]:
        """Get all files in the database.

        Returns:
            List[File]: A list of every file currently stored in the DB.
        """

    @abstractmethod
    def get_file_by_path(self, path: Path) -> Union[File, None]:
        """Get a file object by its path.

        Args:
            path (Path): The Path of the file that you want to get.

        Returns:
            Union[File, None]: The file object if present, None if that file is not in the DB.
        """

    @abstractmethod
    def get_files_by_type(self, filetype: FileType) -> List[File]:
        """Get all files of the specified FileType.

        Args:
            filetype (FileType): The kind of file you want to get.

        Returns:
            List[File]: A list of all Files of the given type.
        """

    @abstractmethod
    def get_files_needing_processing(self) -> List[File]:
        """Get all files that  have the needs_processing flag set.

        Returns:
            List[File]: A list of all Files that have the needs_processing flag set.
        """

    @abstractmethod
    def get_deleted_files(self) -> List[File]:
        """Get all files that have the was_deleted flag set.

        Returns:
            List[File]: A list of all Files that have the was_deleted flag set.
        """

    @abstractmethod
    def remove_file(self, file: File) -> None:
        """Remove this file from the DB.

        Args:
            file (File): The file to remove.
        """

    @abstractmethod
    def add_or_update_file(self, file: File) -> None:
        """Insert or update a file in the DB.

        Args:
            file (File): The file to add or update.
        """


class SQLiteLibrary(LibraryDB):
    _FILES_TABLE = "Files"
    _FILES_COLUMNS = {
        "path": "TEXT PRIMARY KEY",
        "filetype": "INT",
        "mtime": "INT",
        "needs_processing": "BOOLEAN",
        "was_deleted": "BOOLEAN"
    }

    def __init__(self, path: Path, delete: bool = False, pretend: bool = False) -> None:
        self.path = path

        if not delete and pretend:
            # Make a copy of the existing DB and save it to a in-memory DB
            self._init_db_con(pretend=False)
            files = self.get_all_files()
            self._init_db_con(pretend=True)
            for file in files:
                self.add_or_update_file(file)
        elif not delete and not pretend:
            # Default - just init the database
            self._init_db_con(pretend=False)
        elif delete and pretend:
            # Don't actually delete the database, just return an in-memory DB
            self._init_db_con(pretend=True)
        elif delete and not pretend:
            try:
                path.unlink()
                logger.info(f"Removed previous database at {self.path}")
            except FileNotFoundError:
                pass
            except OSError as e:
                logger.fatal((
                    f"Could not remove databse at {self.path}. "
                    f"Please delete the file manually to re-initialize the library. Error: {repr(e)}"
                ))
                raise e
            self._init_db_con()

        logger.debug(f"Initialzed SQLite3 database at {self.path}")

    def get_all_files(self) -> List[File]:
        fetched = self._make_query(f"SELECT * FROM {SQLiteLibrary._FILES_TABLE}")
        return [self._file_from_row(row) for row in fetched]

    def get_file_by_path(self, path: Path) -> Union[File, None]:
        fetched = self._make_query(f"SELECT * FROM {SQLiteLibrary._FILES_TABLE} WHERE path=?", (str(path),))
        if fetched:
            return self._file_from_row(fetched[0])
        else:
            return None

    def get_files_by_type(self, filetype: FileType):
        fetched = self._make_query(f"SELECT * FROM {SQLiteLibrary._FILES_TABLE} WHERE filetype=?", (filetype.value,))
        return [self._file_from_row(row) for row in fetched]

    def get_files_needing_processing(self):
        fetched = self._make_query(f"SELECT * FROM {SQLiteLibrary._FILES_TABLE} WHERE needs_processing")
        return [self._file_from_row(row) for row in fetched]

    def get_deleted_files(self):
        fetched = self._make_query(f"SELECT * FROM {SQLiteLibrary._FILES_TABLE} WHERE was_deleted")
        return [self._file_from_row(row) for row in fetched]

    def add_or_update_file(self, file: File) -> None:
        self._make_query((
            f"INSERT OR REPLACE INTO {SQLiteLibrary._FILES_TABLE} "
            "(path, filetype, mtime, needs_processing, was_deleted) VALUES (?,?,?,?,?)"
        ), (str(file.path), file.type.value, file.mtime, file.needs_processing, file.was_deleted))

    def remove_file(self, file: File) -> None:
        self._make_query(f"DELETE FROM {SQLiteLibrary._FILES_TABLE} WHERE path=?", (str(file.path),))

    def _make_query(self, query: str, params: Union[Dict, Tuple] = ()) -> Union[List, None]:
        """Perform a SQLite query with the given parameters.

        Args:
            query (str): [description]
            params (Union[Dict, Tuple], optional): Parameters in a format supported by sqlite3s execute().
                Defaults to an empty tuple.

        Raises:
            sqlite3.Error: If the query failed

        Returns:
            Union[List, None]: The returned rows. Might be None
        """
        try:
            with self._con:
                logger.debug(f"Running database query '{query}' with parameters {params}")
                result = self._con.execute(query, params).fetchall()
                return result
        except sqlite3.Error as e:
            logger.fatal(f"Error performing database query: {repr(e)}")
            raise e

    def _init_db_con(self, pretend: bool = False):
        """Initialize a datbase connection and ensure the required tables exist.

        Args:
            pretend (bool, optional): Whether the initialize the connection against a temporary in-memory DB.
                Defaults to False
        """
        columns = ', '.join([
            f"{column} {SQLiteLibrary._FILES_COLUMNS[column]}"
            for column in SQLiteLibrary._FILES_COLUMNS
        ])
        try:
            if pretend:
                self._con = sqlite3.connect(":memory:", check_same_thread=False)
            else:
                self.path.parent.mkdir(parents=True, exist_ok=True)
                # Python <= 3.7 needs a string instead of a Path object
                self._con = sqlite3.connect(str(self.path), check_same_thread=False)
            self._con.row_factory = sqlite3.Row
            with self._con:
                self._con.execute(f"CREATE TABLE IF NOT EXISTS {SQLiteLibrary._FILES_TABLE} ({columns})")
        except (sqlite3.Error, OSError) as e:
            logger.fatal(f"Error while accessing/initializing SQLite3 database at {self.path}: {repr(e)}")
            raise e

    @staticmethod
    def _file_from_row(row: Row) -> FileType:
        """Convert a row back into a full file object, including enums and Paths.
        """
        file = File(**row)
        file.path = Path(row["path"])
        file.type = FileType(row["filetype"])
        file.needs_processing = bool(row["needs_processing"])
        file.was_deleted = bool(row["was_deleted"])
        return file


def init(config: Dict, delete: bool = False, pretend: bool = False, exit_on_error: bool = False) -> LibraryDB:
    """Initialize the database connection based on the values provided in the config.

    Creates an apropiate LibraryDB object for the given backend in config and returns it.

    Args:
        config (Dict): MusicBirds configuration
        delete (bool, optional): Whether to re-initialize the DB and remove all previous records. Defaults to False.
        pretend (bool, optional): Whether to use a temporary database that
            does not store writes persistently for pretend mode. Defaults to False.
        exit_on_error (bool, optional): Whether to call sys.exit() if the initialization fails.
            If set to false, will raise exception instead. Defaults to False.

    Returns:
        LibraryDB: The library object
    """
    try:
        if config["database"] == "sqlite3":
            return SQLiteLibrary(config["sqlite3"]["path"], delete, pretend)
    except (OSError, sqlite3.Error) as e:
        if exit_on_error:
            sys.exit("Error initializing database")
        else:
            raise e



"""Interfaces for encoding audio files.

This module provides the interface used by the other MusicBird components to encode audio files.
The Abstract class Encoder serves as the interface used by other components, while individual Encoder implementations
(such as MP3Encoder) are derived from it.
"""
from abc import ABC, abstractmethod
import logging
from pathlib import Path
import subprocess
import re
import sys
from typing import Dict

import ffmpeg

logger = logging.getLogger(__name__)


class Encoder(ABC):
    """Interface for interacting with an Audio file Encoder.

    All encode operations are intended to be done using the methods defined by this class
    """

    extension = ""

    @abstractmethod
    def encode(self, src: Path, dest: Path) -> bool:
        """Encode the file at src to dest.

        Calls the encoders backend and converts the source file, then saves the output at the destination directory.

        Args:
            src (Path): The file to encode.
            dest (Path): The Path at which to store the encoded file.

        Returns:
            bool: True if the operation was successful, False if not.
        """

    @staticmethod
    def mkdir(dest: Path) -> bool:
        """Create the directory for the dest file.

        Returns:
            bool: Whether the creatin was successful
        """
        try:
            dest.parent.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            logger.error(f"Could not create directory to encode file {dest}: {repr(e)}")
            return False
        return True


class FFmpegEncoder(Encoder):
    """Base class for all encoders utilizing ffmpeg.

    Child encoders can inherit from this class and set their parameters accordingly.
    """
    # We might need to use config in this class in the future, so keep it in
    # pylint: disable=unused-argument

    def __init__(self, config: Dict) -> None:
        super().__init__()
        self.ffmpeg_args = {}

        # Check if we can access ffmpeg
        try:
            subprocess.run(["ffmpeg", "-version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        except FileNotFoundError as e:
            logger.fatal("Could not access ffmpeg. Pease make sure that it is installed")
            raise e
        except subprocess.CalledProcessError as e:
            logger.fatal(f"Could not verify that ffmpeg is ready. Error: {repr(e)}")
            raise e

    def encode(self, src: Path, dest: Path) -> bool:
        if not self.mkdir(dest):
            return False
        try:
            logger.debug(f"Encoding to {str(dest)} with arguments {self.ffmpeg_args}")
            stream = ffmpeg.input(str(src))
            stream = ffmpeg.output(stream, str(dest), **self.ffmpeg_args)
            stream = ffmpeg.overwrite_output(stream)
            ffmpeg.run(stream, quiet=True)
        except ffmpeg.Error as e:
            logger.error(f"Failed to encode file {src}. ffmpeg error: \n {e.stderr}")
            return False
        return True


class MP3Encoder(FFmpegEncoder):
    def __init__(self, config: Dict) -> None:
        super().__init__(config)
        self.extension = ".mp3"
        self.ffmpeg_args["acodec"] = "libmp3lame"
        if config["vbr"]:
            self.ffmpeg_args["q:a"] = config["quality"]
        else:
            self.ffmpeg_args["audio_bitrate"] = config["bitrate"]


class OpusEncoder(FFmpegEncoder):

    opusenc_checked = False
    use_opusenc = False

    def __init__(self, config: Dict) -> None:
        super().__init__(config)

        self.extension = ".opus"

        if self._init_opusenc():
            self.opus_args = [
                "--bitrate", re.sub(r'\D', '', config["bitrate"]),  # Strip k postfix from bitrate
            ]
        else:
            # FFmpeg fallback
            self.ffmpeg_args["acodec"] = "libopus"
            self.ffmpeg_args["audio_bitrate"] = config["bitrate"]

    def encode(self, src: Path, dest: Path) -> bool:
        if self.use_opusenc:
            if not self.mkdir(dest):
                return False

            args = ["opusenc", str(src), str(dest.with_suffix(self.extension))] + self.opus_args
            try:
                subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
            except (subprocess.CalledProcessError, OSError) as e:
                logger.error(f"fFailed to encode file {src}. opusenc error: \n {repr(e)}")
                return False
            return True
        else:
            return super().encode(src, dest)

    def _init_opusenc(self) -> bool:
        """Look for opusenc and set the encoder to use it if available.

        FFmpegs libopus encoder does not support embedding album art as of late 2021.
        To work around this, we use opusenc directly if it is installed. If not,
        we issue a warning and fallback to ffmpeg.

        Returns:
            bool: Whether opusenc can be used.
        """

        # Use a classvar to only check for opusenc once every run.
        # If it's not there the first, it's not gonna be there at all.
        if OpusEncoder.opusenc_checked:
            return OpusEncoder.use_opusenc

        try:
            subprocess.run(["opusenc", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        except (FileNotFoundError, subprocess.CalledProcessError):
            logger.warning((
                "opusenc does not appear to be installed. Falling back to ffmpeg encoding."
                "Please install 'opus-tools' if you want embedded album art support in Opus files"
            ))
            OpusEncoder.use_opusenc = False
        else:
            OpusEncoder.use_opusenc = True
        finally:
            OpusEncoder.opusenc_checked = True
        return OpusEncoder.use_opusenc


def init(config, exit_on_error: bool = True) -> Encoder:
    """Initialize an encoder object based on the values provided in the config.

    Creates an apropiate LibraryDB object for the given backend in config and returns it.

    Args:
        config (Dict): MusicBirds configuration
        exit_on_error (bool, optional): Whether to call sys.exit() if the initialization fails.
            If set to false, will raise exception instead. Defaults to False.

    Returns:
        Encoder: The Encoder object
    """
    try:
        if config["encoder"] == "mp3":
            return MP3Encoder(config["mp3"])
        elif config["encoder"] == "opus":
            return OpusEncoder(config["opus"])
    except (OSError, ffmpeg.Error) as e:
        if exit_on_error:
            sys.exit("Error initializing database")
        else:
            raise e



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['musicbird']

package_data = \
{'': ['*'], 'musicbird': ['data/*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'ffmpeg-python>=0.2.0,<0.3.0', 'schema>=0.7.5,<0.8.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0'],
 'docs': ['Sphinx>=4.3.2,<5.0.0',
          'sphinx-rtd-theme>=1.0.0,<2.0.0',
          'sphinxcontrib-napoleon>=0.7,<0.8']}

entry_points = \
{'console_scripts': ['musicbird = musicbird.__main__:entrypoint']}

setup_kwargs = {
    'name': 'musicbird',
    'version': '0.1.8',
    'description': 'Convert your music library on-the-fly!',
    'long_description': "MusicBird - Convert your Music library on-the-fly 🐦\n####################################################\n\n.. image:: https://img.shields.io/github/workflow/status/maxhoesel/MusicBird/CI.svg\n   :target: https://img.shields.io/github/workflow/status/maxhoesel/MusicBird/CI.svg\n.. image:: https://img.shields.io/pypi/pyversions/musicbird.svg\n   :target: https://img.shields.io/pypi/pyversions/musicbird.svg\n.. image:: https://img.shields.io/pypi/l/musicbird.svg\n   :target: https://img.shields.io/pypi/l/musicbird.svg\n.. image:: https://img.shields.io/codecov/c/github/maxhoesel/MusicBird.svg\n   :target: https://img.shields.io/codecov/c/github/maxhoesel/MusicBird.svg\n\n----\n\nMusicBird is a python package that creates a mobile-friendly copy of your music library. Its major features include:\n\n* It's *fast*! Not only does it use all your cores when encoding files, it also remembers the state of your library from when you last ran it.\n  This means that MusicBird will only process those files that have actually changed since then.\n* It works with any music library! You don't need to adjust your library structure at all for MusicBird to do its magic - in fact, you could make it read-only\n  and everything would still work just fine!\n* It tracks modified and deleted files! Did you change the tags of a file and want them on your phone as well?\n  Do you no longer like that artist and deleted all their music from your library? Don't worry!\n  MusicBird will pick up those changes and adjust the mirror copy accordingly\n* It's flexible! MusicBird uses the excellent `ffmpeg libraries <https://ffmpeg.org/>`_ under the hood,\n  meaning that most common input and output formats are supported.\n\nDocumentation\n=============\n\nSee the `official docs <https://musicbird.readthedocs.io/en/latest/>`_ for installation and usage instructions\n\nAuthor & License\n================\n\nWritten and maintained by Max Hösel (@maxhoesel)\n\nLicensed under the GNU GPLv3 or later\n",
    'author': 'Max Hösel',
    'author_email': 'musicbird@maxhoesel.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/maxhoesel/musicbird',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<3.10',
}


setup(**setup_kwargs)



"""Provides the run command and related functions.
"""

import argparse
import logging
from typing import Dict, List

from .db import init as init_db
from .scan import scan
from .copy import copy
from .encode import encode
from .prune import prune

logger = logging.getLogger(__name__)


def run_command(parent_parser: argparse.ArgumentParser, args: List[str], config: Dict):
    """Entrypoint for the CLI `run` command.

    Args:
        parent_parser (argparse.ArgumentParser): The parser from the main entrypoint.
            Used to display a full --help output by inheriting its arguments.
        args (List[str]): List of arguments not parsed by the main parser.
        config (Dict): Dictionary containing the MusicBird configuration
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=__doc__, prog="musicbird", parents=[parent_parser])
    parser.add_argument("--rescan", action="store_true",
                        help="Force a rescan of the entire library. This will force all files to be reprocessed")
    parser.add_argument("--pretend", action="store_true",
                        help="Show what changes would be made but don't modify any files")
    args = parser.parse_args(args)
    return run(config, args.rescan, args.pretend)


def run(config: Dict, rescan: bool = False, pretend: bool = False):
    """Scan, then process the entire music library.

    Equivalent to calling scan(), copy(), encode(), prune() in that order.

    Args:
        config (Dict): Dictionary containing the musicbird configuration.
        db (LibraryDB): Database object to read/write the library status from/to.
        pretend (bool, optional): Pretend to process/scan, but don't perform any actual operations. Defaults to False.

    Returns:
        bool: True if all files were processed successfully, false if not.
    """
    db = init_db(config, delete=rescan, pretend=pretend)

    results = []
    results.append(scan(config, db))
    results.append(copy(config, db, pretend=pretend))
    results.append(encode(config, db, pretend=pretend))
    results.append(prune(config, db, pretend=pretend))
    if False in results:
        return False
    else:
        return True
